/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: NetBuffer.cpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#include "NetBuffer.hpp"
#include "logging.hpp"
#include "Cipher.hpp"


EPS::NetBuffer::NetBuffer(const EPS::NetBuffer::EType type) :
	length		( 0			),
	related		( nullptr	),
	buffer_type	( type		),
	buffer_1st	( &buffer[ 0						], buffer.max_size()							),
	buffer_2nd	( &buffer[ EPS::cipher_block_size	], buffer.max_size() - EPS::cipher_block_size	)
{
	clear(type);

	return;
}


EPS::NetBuffer::NetBuffer(const NetBuffer &nb) :
	NetBuffer(nb.buffer_type)
{
	this->clone(nb);

	return;
}


EPS::NetBuffer::~NetBuffer(void)
{
//	LOG_TRACE << "~NetBuffer " << (void*)this << ": " << name() << " of length=" << length << " bytes is being deleted";

	if (this->buffer_type == EPS::NetBuffer::EType::kInvalid || length > 2000)
	{
		LOG_ERROR << "invalid buffer " << (void*)this << ": " << name() << " with a length of " << length << " bytes; ignoring \"related\" field";
	}
	else if (this->related)
	{
//		LOG_TRACE << "related " << name(this->related->buffer_type) << " buffer " << (void*)this->related << " will also be deleted";
		this->related->related = nullptr; // prevent recursion in the destructor
		delete this->related;
//		LOG_TRACE << "related buffer has been deleted";
	}

//	LOG_TRACE << "~NetBuffer " << (void*)this << ": " << name() << " of length=" << length << " bytes has been deleted";

	this->buffer_type	= EPS::NetBuffer::EType::kInvalid;
	this->length		= size_t(-1);
	this->related		= nullptr;

	return;
}


void EPS::NetBuffer::clear(const bool also_clear_related_buffer)
{
	// make sure the buffer cannot be mistaken for a valid buffer by wiping out the first few byes of the header
	memset(this->buffer.data(), '\0', EPS::cipher_block_size);

	this->length		= 0;
	this->buffer_type	= EPS::NetBuffer::EType::kInvalid;

	if (also_clear_related_buffer && this->related)
	{
		related->clear(false);
	}

	return;
}


void EPS::NetBuffer::clear(const EPS::NetBuffer::EType type)
{
	this->clear(true);
	this->buffer_type = type;

	return;
}


EPS::NetBuffer *EPS::NetBuffer::get(const EPS::NetBuffer::EType type)
{
	// if this is what we already have, then we have nothing remaining to do
	if (type == this->buffer_type)
	{
		return this;
	}

	// handle the special exception case: if all we want is the header, then a fully-decrypted buffer is valid
	if (type 				== EPS::NetBuffer::EType::kPartiallyDecrypted	&&
		this->buffer_type	== EPS::NetBuffer::EType::kFullyDecrypted		)
	{
		return this;
	}

	// if the related pointer is what was requested, then we have nothing remaining to do
	if (this->related)
	{
		const EPS::NetBuffer::EType related_buffer_type = this->related->buffer_type;

		if (type == related_buffer_type)
		{
			return this->related;
		}

		// handle the special exception case: if all we want is the header, then a fully-decrypted buffer is also valid
		if (type 				== EPS::NetBuffer::EType::kPartiallyDecrypted	&&
			related_buffer_type	== EPS::NetBuffer::EType::kFullyDecrypted		)
		{
			return this->related;
		}
	}

	// if we get here, then we need to either encrypt or decrypt a buffer

	switch (type)
	{
		case EPS::NetBuffer::EType::kPartiallyDecrypted:
		{
			// decrypt just the header
			return EPS::Cipher::decrypt_header(*this);
		}

		case EPS::NetBuffer::EType::kFullyDecrypted:
		{
			// decrypt the entire buffer
			return EPS::Cipher::decrypt(*this);
		}

		case EPS::NetBuffer::EType::kEncrypted:
		{
			// encrypt the buffer
			return EPS::Cipher::encrypt(*this);
		}

		default:
		{
			// what in the world have we been asked to do with this buffer?
			break;
		}
	}

	THROW_INVALID_ARGUMENT(-1, "Cannot convert network buffer from " + this->name() + " to " + this->name(type) + ".");
}


bool EPS::NetBuffer::lengths_are_valid(void)
{
	EPS::NetBuffer *encrypted_buffer = nullptr;
	EPS::NetBuffer *decrypted_buffer = nullptr;

	EPS::NetBuffer::EType type = this->buffer_type;
	if		(type  == EPS::NetBuffer::EType::kEncrypted)			encrypted_buffer = this;
	else if	(type  == EPS::NetBuffer::EType::kFullyDecrypted)		decrypted_buffer = this;
	else if	(type  == EPS::NetBuffer::EType::kPartiallyDecrypted)	decrypted_buffer = this;

	EPS::NetBuffer::EType rtype	= EPS::NetBuffer::EType::kInvalid;
	if (this->related)
	{
		rtype = this->related->buffer_type;
		if		(rtype == EPS::NetBuffer::EType::kEncrypted)			encrypted_buffer = this->related;
		else if (rtype == EPS::NetBuffer::EType::kFullyDecrypted)		decrypted_buffer = this->related;
		else if (rtype == EPS::NetBuffer::EType::kPartiallyDecrypted)	decrypted_buffer = this->related;
	}

	if (encrypted_buffer == nullptr && decrypted_buffer == nullptr)
	{
		// nothing we can validate given these buffers
		LOG_ERROR << "cannot validate buffer lengths given buffer types of " << this->name(type) << " and " << this->name(rtype);
		return false;
	}

	size_t encryption_length		= 0;
	size_t decryption_length		= 0;
	size_t decrypted_header_length	= 0;
	size_t fully_padded_length		= 0;

	if (encrypted_buffer)
	{
		encryption_length	= encrypted_buffer->length;
		fully_padded_length	= encryption_length;

		if (encryption_length < EPS::cipher_block_size		||
			encryption_length > EPS::network_buffer_size	)
		{
			LOG_ERROR << "encrypted buffer of length " << encryption_length << " cannot possibly be valid";
			return false;
		}

		if ((encryption_length % EPS::cipher_block_size) != 0)
		{
			LOG_ERROR << "encrypted buffer of length " << encryption_length << " is not a multiple of " << EPS::cipher_block_size << " bytes";
			return false;
		}
	}

	if (decrypted_buffer)
	{
		decryption_length			= decrypted_buffer->length;
		EPS::CommHeader1 *header	= decrypted_buffer->get_header();
		decrypted_header_length		= header->size_of_packet();
		fully_padded_length			= EPS::length_including_pkcs_padding(decrypted_header_length);

		if (decryption_length < EPS::cipher_block_size)
		{
			LOG_ERROR << "decrypted buffer of length " << decryption_length << " is too small to be valid";
			return false;
		}

		// is this a fully-decrypted, or partially-decrypted packet?

		if (decrypted_buffer->is_fully_decrypted())
		{
			if (decryption_length != decrypted_header_length)
			{
				LOG_ERROR << "decrypted buffer contains " << decryption_length << " bytes, but header indicates we should have " << decrypted_header_length << " bytes";
				return false;
			}

			if (decryption_length + EPS::cipher_block_size > EPS::network_buffer_size)
			{
				LOG_ERROR << "decrypted buffer of length " << decryption_length << " is too large to be valid";
				return false;
			}
		}
		else
		{
			// the packet hasn't yet been fully decrypted

			if (fully_padded_length != decryption_length)
			{
				LOG_ERROR << "partially decrypted buffer should contain " << fully_padded_length << " bytes, but instead contains " << decryption_length << " bytes";
				return false;
			}

			if (decryption_length > EPS::network_buffer_size)
			{
				LOG_ERROR << "partially decrypted buffer of length " << decryption_length << " is too large to be valid";
				return false;
			}
		}
	}

	// if we have both the encrypted and decrypted packets, then there are a few more tests we can run between them
	if (encrypted_buffer && decrypted_buffer)
	{
		if (encryption_length < decryption_length							||
			encryption_length > decryption_length + EPS::cipher_block_size	)
		{
			LOG_ERROR << "encrypted length of " << encryption_length << " bytes and decrypted length of " << decryption_length << " bytes cannot possibly be valid";
			return false;
		}

		if (encryption_length != fully_padded_length)
		{
			LOG_ERROR << "encryption length of " << encryption_length << " differs from the expected fully padded length of " << fully_padded_length << " bytes";
			return false;
		}
	}

	// if we get here, then it looks like the buffer sizes are good
	return true;
}


std::string EPS::NetBuffer::name(void) const
{
	return this->name(this->buffer_type);
}


std::string EPS::NetBuffer::name(const EPS::NetBuffer::EType type) const
{
	return	type == EPS::NetBuffer::EType::kInvalid				? "invalid buffer"				:
			type == EPS::NetBuffer::EType::kEncrypted 			? "encrypted buffer"			:
			type == EPS::NetBuffer::EType::kPartiallyDecrypted	? "partially decrypted buffer"	:
			type == EPS::NetBuffer::EType::kFullyDecrypted		? "fully decrypted buffer"		: "unknown buffer";
}


EPS::NetBuffer *EPS::NetBuffer::create_related_buffer(const EPS::NetBuffer::EType type)
{
	// get a new "related" buffer, or completely reset the one we already have

	if (type == this->buffer_type)
	{
		THROW_INVALID_ARGUMENT(-1, "buffer is already the same as the requested type: " + this->name());
	}

	if (this->related)
	{
		this->related->clear(false);
	}
	else
	{
		this->related = new EPS::NetBuffer;
	}

	this->related->buffer_type	= type;
	this->related->related		= this;

	return this->related;
}


void EPS::NetBuffer::clone(const EPS::NetBuffer &nb)
{
	if (this == &nb)
	{
		return;
	}

	this->clear();
	this->buffer_type	= nb.buffer_type;
	this->length		= nb.length;
	memcpy(this->data(), nb.data(), nb.length);

	return;
}
