/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: TCPConnection.cpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#include "TCPConnection.hpp"
#include "Hull.hpp"
#include "logging.hpp"
#include "common.hpp"
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <sstream>


EPS::TCPConnection::TCPConnection(void) :
	is_mesh(false),
	socket(hull.io_service)
{
	/** Create a brand new TCP connection object but don't automatically connect the socket anywhere.
	 * For example, this is used when the listener needs an object to accept a new connection.
	 * Expect @ref EPS::TCPConnection::accepted() to be called once an actual connection has been accepted.
	 */

	this->name = std::to_string(this->id) + "->? ";

	return;
}


EPS::TCPConnection::TCPConnection(const EPS::TCPEP &eth_endpoint, const EPS::UUID &uuid) :
	is_mesh(false),
	socket(hull.io_service),
	endpoint(eth_endpoint)
{
	/// Create a brand new TCP connection to the endpoint provided.

	if (not uuid.is_nil())
	{
		// if we know the UUID before we've established a connection, then it means we're establishing a mesh connection
		this->uuid		= uuid;
		this->is_mesh	= true;
	}

	this->name = std::to_string(this->id) + "->" + this->endpoint.address().to_string() + ":" + std::to_string(this->endpoint.port()) + ": ";

	return;
}


EPS::TCPConnection::~TCPConnection(void)
{
	if (socket.is_open())
	{
		LOG_TRACE << this->name << "closing open connection to " << this->endpoint;

		socket.cancel();
		LOG_TRACE << this->name << "socket has been cancelled";

		socket.close(this->ec);
		LOG_TRACE << this->name << "socket has been closed";
	}

	// get rid of any pending writes which remain in the deque
	if (this->pending_write_deque.empty() == false)
	{
		LOG_TRACE << this->name << "need to delete " << this->pending_write_deque.size() << " pending writes";

		for (EPS::NetBuffer *nb : this->pending_write_deque)
		{
			delete nb;
		}
		this->pending_write_deque.clear();

		LOG_TRACE << this->name << "pending write deque is now empty";
	}

	LOG_TRACE << this->name << "destructing connection";

	return;
}


void EPS::TCPConnection::async_connect(void)
{
	if (this->endpoint.address().is_unspecified())
	{
		THROW_INVALID_ARGUMENT(-1, "Cannot connect to unspecified IP address.");
	}
	if (this->endpoint.port() < 1)
	{
		THROW_INVALID_ARGUMENT(-1, "Cannot connect to invalid TCP port " + std::to_string(this->endpoint.port()) + ".");
	}
	if (this->socket.is_open())
	{
		THROW_RUNTIME_ERROR(-1, "Socket is already open.");
	}

	this->socket.open( endpoint.protocol(), this->ec );
	this->set_socket_options();

	// attempt to connect to the endpoint provided when the object was first constructed
	LOG_TRACE << this->name << "connecting to " << this->endpoint;

	this->socket.async_connect( this->endpoint,
			[=](const EC &error)
			{
				this->ec = error;
				if (error)
				{
					LOG_ERROR << name << "tcp beam to " << this->endpoint << " has failed: " << error.message();
					hull.connection_failed(this);
				}
				else
				{
					LOG_DEBUG << name << "tcp beam established to " << this->endpoint;
					hull.validate_new_connection(this);
				}
				return;
			} );

	return;
}


void EPS::TCPConnection::accepted(void)
{
	/// Called after a new TCP connection has been accepted by a listener/acceptor.

	this->endpoint	= this->socket.remote_endpoint(this->ec);
	this->name		= std::to_string(this->id) + "->" + this->endpoint.address().to_string() + ":" + std::to_string(this->endpoint.port()) + ": ";

	this->set_socket_options();

	LOG_INFO << name << "connection accepted";

	return;
}


void EPS::TCPConnection::set_socket_options(void)
{
	socket.non_blocking													(false						, ec );
	socket.native_non_blocking											(false						, ec );

	socket.set_option( boost::asio::ip::tcp::no_delay					(true)						, ec );
	socket.set_option( boost::asio::socket_base::linger					(false, 0)					, ec );
	socket.set_option( boost::asio::socket_base::keep_alive				(true)						, ec );
	socket.set_option( boost::asio::socket_base::reuse_address			(true)						, ec );
	socket.set_option( boost::asio::socket_base::send_buffer_size		(EPS::network_buffer_size)	, ec );
	socket.set_option( boost::asio::socket_base::receive_buffer_size	(EPS::network_buffer_size)	, ec );

	// Reads are handled async, but writes are expected to complete immediately.
	// We want to prevent somehow blocking forever on a call to write.

#if 0
	struct timeval timeout;
	timeout.tv_sec	= 2;
	timeout.tv_usec	= 0;

	setsockopt(socket.native_handle(), SOL_SOCKET, SO_SNDTIMEO, &timeout, sizeof(timeout));
#endif

	return;
}
