/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: ServerUpnp.cpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#include "ServerHull.hpp"
#include "logging.hpp"

#ifndef EPS_SKIP_UPNP

#include <miniupnpc/miniupnpc.h>
#include <miniupnpc/upnpcommands.h>
#include <miniupnpc/upnperrors.h>


/// why-oh-why are the UPnP error codes not defined in miniupnp? @see @p strupnperror() @{
#define UPNP_ERROR_INVALID_ACTION						401
#define UPNP_ERROR_INVALID_ARGS							402
#define UPNP_ERROR_ACTION_FAILED						501
#define UPNP_ERROR_ACTION_NOT_AUTHORIZED				606
#define UPNP_ERROR_PINHOLE_SPACE_EXHAUSTED				701
#define UPNP_ERROR_FIREWALL_DISABLED					702
#define UPNP_ERROR_INBOUND_PINHOLE_NOT_ALLOWED			703
#define UPNP_ERROR_NO_SUCH_ENTRY						704
#define UPNP_ERROR_PROTOCOL_NOT_SUPPORTED				705
#define UPNP_ERROR_INTERNAL_PORT_WILDCARD_NOT_ALLOWED	706
#define UPNP_ERROR_PROTOCOL_WILDCARD_NOT_ALLOWED		707
#define UPNP_ERROR_SRC_IP_WILDCARD_NOT_ALLOWED			708
#define UPNP_ERROR_NO_PACKET_SENT						709
#define UPNP_ERROR_ARRAY_INDEX_INVALID					713
#define UPNP_ERROR_NO_SUCH_ENTRY_IN_ARRAY				714
#define UPNP_ERROR_SRC_IP_WILCARD_NOT_PERMITTED			715
#define UPNP_ERROR_EXT_PORT_WILDCARD_NOT_PERMITTED		716
#define UPNP_ERROR_CONFLICT_IN_MAP_ENTRY				718
#define UPNP_ERROR_SAME_PORT_VALUES_REQUIRED			724
#define UPNP_ERROR_ONLY_PERMANENT_LEASE_SUPPORTED		725
#define UPNP_ERROR_REMOTE_HOST_ONLY_SUPPORTS_WILDCARD	726
#define UPNP_ERROR_EXT_PORT_ONLY_SUPPORTS_WILDCARD		727
///}

#define SECONDS_IN_24_HOURS	(60 * 60 * 24)


bool tcp_port_mapping_needs_to_be_deleted = false;
bool udp_port_mapping_needs_to_be_deleted = false;
bool tcp_port_mapping_needs_to_be_created = false;
bool udp_port_mapping_needs_to_be_created = false;


void EPS::ServerHull::setup_UPnP(void)
{
	if (not this->upnp_enabled)
	{
		LOG_INFO << "UPnP support is disabled; skipping";
		return;
	}

	if (this->upnp_timeout < 250)
	{
		LOG_WARN << "UPnP timeout is too short; skipping";
		this->upnp_enabled = false;
		return;
	}

	int error = 0;
	this->upnp_dev = upnpDiscover(
			upnp_timeout,	// time to wait (milliseconds)
			nullptr		,	// multicast interface (or null defaults to 239.255.255.250)
			nullptr		,	// path to minissdpd socket (or null defaults to /var/run/minissdpd.sock)
			0			,	// source port to use (or default port 1900)
			0			,	// 0==IPv4, 1==IPv6
			&error		);	// error condition

	if (this->upnp_dev == nullptr || error != 0)
	{
		LOG_WARN << "UPnP support not detected (waited " << upnp_timeout << " milliseconds, error #" << error << ": " << strupnperror(error) << ")";
		this->upnp_enabled = false;
		return;
	}

	std::string lan_address; // our own ip address on the LAN
	std::string wan_address; // router/nat address on the WAN
	char tmp_buffer[64];

	error = UPNP_GetValidIGD(this->upnp_dev, &this->upnp_urls, &this->upnp_data, tmp_buffer, sizeof(tmp_buffer));
	/*	-1	== Internal error
	 *	0	== NO IGD found
	 *	1	== A valid connected IGD has been found
	 *	2	== A valid IGD has been found but it reported as not connected
	 *	3	== an UPnP device has been found but was not recognized as an IGD
	 */
	if (error != 1)
	{
		LOG_WARN << "UPnP: IGD state = " << error << "; disabling UPnP";
		this->upnp_enabled = false;
		return;
	}

	lan_address = tmp_buffer;
	this->upnp_lan_address = EPS::IPv4::from_string(lan_address, this->ec);
	LOG_INFO << "UPnP: LAN address: " << lan_address;

	error = UPNP_GetExternalIPAddress(this->upnp_urls.controlURL, this->upnp_data.first.servicetype, tmp_buffer);
	wan_address = tmp_buffer;
	this->upnp_wan_address = EPS::IPv4::from_string(wan_address, this->ec);
	LOG_INFO << "UPnP: WAN address: " << wan_address;

	this->renew_UPnP_mappings();

	return;
}


void EPS::ServerHull::reset_UPnP(void)
{
	if (this->upnp_enabled)
	{
		if (tcp_port_mapping_needs_to_be_deleted) delete_UPnP_mapping("TCP");
		if (udp_port_mapping_needs_to_be_deleted) delete_UPnP_mapping("UDP");
	}

	FreeUPNPUrls	( &this->upnp_urls	);
	freeUPNPDevlist	(  this->upnp_dev	);

	return;
}


void EPS::ServerHull::renew_UPnP_mappings(void)
{
	long unsigned int time_to_sleep_in_seconds = SECONDS_IN_24_HOURS + 15;

	if (this->upnp_enabled)
	{
		LOG_INFO << "UPnP: renewing port mappings";

		this->read_UPnP_mappings();

		// delete old port mappings
		if (tcp_port_mapping_needs_to_be_deleted) this->delete_UPnP_mapping("TCP");
		if (udp_port_mapping_needs_to_be_deleted) this->delete_UPnP_mapping("UDP");

		// create new port mappings
		bool success = true;
		if (tcp_port_mapping_needs_to_be_created) success &= this->create_UPnP_mapping("TCP");
		if (udp_port_mapping_needs_to_be_created) success &= this->create_UPnP_mapping("UDP");

		if (not success)
		{
			// something went wrong and we failed to set the UPnP mapping, so try again soon
			LOG_WARN << "UPnP: failed to renew the port mappings, so shortening the next wait time down to 30 seconds";
			time_to_sleep_in_seconds = 30;
		}
	}

	// re-start the timer to regularly renew the port mappings
	LOG_INFO << "UPnP: setting up a timer to renew the port mappings in " << time_to_sleep_in_seconds << " seconds";
	this->upnp_renew_timer.expires_from_now( std::chrono::seconds(time_to_sleep_in_seconds) );
	this->upnp_renew_timer.async_wait(
			[=](const EC &error)
			{
				if (not error && not this->done)
				{
					this->renew_UPnP_mappings();
				}

				return;
			} );

	return;
}


void EPS::ServerHull::read_UPnP_mappings(void)
{
	if (not this->upnp_enabled)
	{
		// nothing we can do
		return;
	}

	// assume we're going to have to create new port mapping rules
	tcp_port_mapping_needs_to_be_created = true;
	udp_port_mapping_needs_to_be_created = true;
	tcp_port_mapping_needs_to_be_deleted = false;
	udp_port_mapping_needs_to_be_deleted = false;

	const std::string lan_address = this->upnp_lan_address.to_string();

	// see if we have a relevant port mapping
	size_t index = 0;
	while (true)
	{
		char map_wan_port			[200] = "";
		char map_lan_address		[200] = "";
		char map_lan_port			[200] = "";
		char map_protocol			[200] = "";
		char map_description		[200] = "";
		char map_mapping_enabled	[200] = "";
		char map_remote_host		[200] = "";
		char map_lease_duration		[200] = "";

		int error = UPNP_GetGenericPortMappingEntry(
					this->upnp_urls.controlURL			,
					this->upnp_data.first.servicetype	,
					std::to_string(index).c_str()		,
					map_wan_port						,
					map_lan_address						,
					map_lan_port						,
					map_protocol						,
					map_description						,
					map_mapping_enabled					,
					map_remote_host						,
					map_lease_duration					);

		if (error)
		{
			break;
		}

		if ( atoi(map_wan_port) == this->port)
		{
			// we found a mapping for the port we want
			LOG_DEBUG << "UPnP: mapping found for " << map_protocol << " WAN port " << map_wan_port << " to " << map_lan_address << ":" << map_lan_port << "; " << map_lease_duration << " seconds (" << map_description << ")";

			if (lan_address != map_lan_address || this->port != atoi(map_lan_port))
			{
				LOG_WARN << "UPnP: " << map_protocol << " WAN port " << map_wan_port << "->" << map_lan_address << ":" << map_lan_port << " does not map to " << lan_address << ":" << this->port << ". Cannot use or modify UPnP port mapping.";
				tcp_port_mapping_needs_to_be_deleted = false;
				udp_port_mapping_needs_to_be_deleted = false;
				this->upnp_enabled = false;
				return;
			}
			else
			{
				// we already have a port mapping entry -- is it temporary or permanent?
				if (atoi(map_lease_duration) == 0)
				{
					LOG_INFO << "UPnP: permanent " << map_protocol << " port mapping already exists for " << map_wan_port << " -> " << map_lan_address << ":" << map_lan_port << "; no need to modify";
					if (std::string(map_protocol) == "TCP")
					{
						tcp_port_mapping_needs_to_be_deleted = false;
						tcp_port_mapping_needs_to_be_created = false;
					}
					else
					{
						udp_port_mapping_needs_to_be_deleted = false;
						udp_port_mapping_needs_to_be_created = false;
					}
				}
				else
				{
					// we have no idea how much time is left on this lease, so delete it and re-create it
					if (std::string(map_protocol) == "TCP")
					{
						tcp_port_mapping_needs_to_be_deleted = true;
						tcp_port_mapping_needs_to_be_created = true;
					}
					else
					{
						udp_port_mapping_needs_to_be_deleted = true;
						tcp_port_mapping_needs_to_be_created = true;
					}
				}
			}
		}

		index ++;
	}

	return;
}


bool EPS::ServerHull::delete_UPnP_mapping(const std::string &protocol)
{
	/// @return TRUE if deleting the port mapping was successful.
	bool result = true;

	if (this->upnp_enabled)
	{
		const int error = UPNP_DeletePortMapping(
							this->upnp_urls.controlURL,
							this->upnp_data.first.servicetype,
							std::to_string(this->port).c_str()	,	// external (WAN) port number
							protocol.c_str()					,	// TCP or UDP
							""									);	// remote host

		if (error == 0)
		{
			LOG_TRACE << "UPnP: successfully deleted mapping for " << protocol << " port " << this->port;
		}
		else
		{
			result = false;
			LOG_ERROR << "UPnP: failed to delete mapping for " << protocol << " port " << this->port << " (error #" << error << ": " << strupnperror(error) << ")";
		}

		// if we deleted a port mapping rule, then we need to remember to create a new one
		if (protocol == "TCP")
		{
			tcp_port_mapping_needs_to_be_created = true;
			tcp_port_mapping_needs_to_be_deleted = true;
		}
		else
		{
			udp_port_mapping_needs_to_be_created = true;
			udp_port_mapping_needs_to_be_deleted = true;
		}
	}

	return result;
}


bool EPS::ServerHull::create_UPnP_mapping(const std::string &protocol)
{
	/// @return TRUE if adding the port mapping was successful.
	bool result = true;

	if (this->upnp_enabled)
	{
		const std::string description	= "EPS Conduits server for " + this->network_str;
		const std::string port_str		= std::to_string(this->port);

		const int error = UPNP_AddPortMapping(
							this->upnp_urls.controlURL			,
							this->upnp_data.first.servicetype	,
							port_str							.c_str(),	// external (WAN) port requested
							port_str							.c_str(),	// internal (LAN) port to which packets will be redirected
							this->upnp_lan_address.to_string()	.c_str(),	// internal (LAN) address to which packets will be redirected
							description							.c_str(),	// text description to indicate why or who is responsible for the port mapping
							protocol							.c_str(),	// protocol must be either TCP or UDP
							nullptr								,			// remote host address
							std::to_string(SECONDS_IN_24_HOURS)	.c_str()	// lease duration (in seconds) or zero for "as long as possible"
							);
		if (error == 0)
		{
			LOG_DEBUG << "UPnP: " << protocol << " port mapping added for " << this->upnp_wan_address.to_string() << ":" << this->port << " -> " << this->upnp_lan_address.to_string() << ":" << this->port;
		}
		else
		{
			result = false;
			LOG_ERROR << "UPnP: failed to add a " << protocol << " port mapping for " << this->upnp_wan_address.to_string() << ":" << this->port << " -> " << this->upnp_lan_address.to_string() << ":" << this->port << " (error #" << error << ": " << strupnperror(error) << ")";
		}

		// if we've created a port mapping rule, then we need to remember to delete it when we eventually exit
		if (protocol == "TCP")
		{
			tcp_port_mapping_needs_to_be_deleted = true;
		}
		else
		{
			udp_port_mapping_needs_to_be_deleted = true;
		}
	}

	return result;
}

#endif
